--{{{ Libraries

local L = AceLibrary("AceLocale-2.2"):new("GridStatusAurasExt")

--}}}

GridStatusAurasExt = GridStatus:NewModule("GridStatusAurasExt")
GridStatusAurasExt.menuName = L["AurasExt"]


--{{{ AceDB defaults

GridStatusAurasExt.defaultDB = {
	auraGroups = {},
}

--}}}

GridStatusAurasExt.options = false

local options = {
    type = "group",
    name = L["AurasExt"],
    desc = L["General option for Auras Ext."],
    args = {
        ["newgroup"] = {
            type = "text",
            name = L["New aura group"],
            desc = L["Creates a new status for a group of auras."],
            get = function ()
                return ""
                end,
            set = function(v)
                GridStatusAurasExt:NewAuraGroup(v)
            end,
            order = 10,
            usage = "",
        },
        ["remove"] = {
            type = "group",
            name = L["Remove group"],
            desc = L["Removes an aura group."],
            args = {},           
        },
    },
}

Grid.options.args["aurasext"] = options

local settings

local auraList = {}
local groupList = {}

function GridStatusAurasExt:OnInitialize()

	self.super.OnInitialize(self)
    settings = GridStatusAurasExt.db.profile
    
    for name, data in pairs(settings.auraGroups) do
        self:RegisterAuraGroup(name, data)
    end    
end

function GridStatusAurasExt:RemoveAuraGroup(name) 
    local status = "status_"..name
    self:UnregisterStatus(status)
    
    local sets = settings[status]
    for id, _ in pairs(sets.ids) do
        self:UnregisterId(status, id)
    end
    
    for aura, _ in pairs(sets.names) do
        self:UnregisterName(status, aura)
    end
    
    groupList[status] = nil
    settings[status] = nil
    options.args.remove.args[status] = nil
    
    settings.auraGroups[name] = nil
end

function GridStatusAurasExt:RegisterAuraGroup(name, data)
    local status = "status_"..name
    
    local sets = settings[status]
    if not  sets then
        sets = {
            enable = true, 
            color = { r = 1, g = 1, b = 1, a = 1 }, 
            priority = 50,  
            allMissing = false,
            onlyTheirs = false,
            onlyMine = false,
            ids = {},
            names = {},
        }
        settings[status] = sets
    end
    
    options.args.remove.args[status] = {
        type = "execute",
        name = name.." - RELOAD UI!",
        desc = L["Removes this aura group. Please Reload your UI after that!"],        
        func = function() self:RemoveAuraGroup(name) end,
        order = 1,
    }
    
    groupList[status] = {settings = sets}
    
    local opts = {
        ["addbyid"] = {
            type = "text",
            name = L["New aura by ID"],
            desc = L["Adds a new Aura with the passed ID to this group"],
            get = function() return "" end,
            set = function(v) GridStatusAurasExt:NewAuraIDToGroup(status, v) end,
            usage = L["<spell id>"],
            order = 150,
        },
        ["addbyname"] = {
            type = "text",
            name = L["New aura by name"],
            desc = L["Adds a new Aura with the passed name to this group"],
            get = function() return "" end,
            set = function(v) GridStatusAurasExt:NewAuraNameToGroup(status, v) end,
            usage = L["<aura name>"],
            order = 151,
        },
        ["remove"] = {
            type = "group",
            name = L["Remove"],
            desc = L["Remove an aura from the group."],
            args = {},
        },
        ["onlymine"] = {
            type = "toggle",
            name = L["Only mine"],
            desc = L["Activates this aura group only for auras cast by you."],
            get = function()
                return sets.onlyMine
                end,
            set = function(v)
                sets.onlyMine = v
                if v and sets.onlyTheirs then
                    sets.onlyTheirs = false
                end
                self:UpdateAllUnitAuras()
            end,
            order = 142,
        },
        ["onlytheirs"] = {
            type = "toggle",
            name = L["Only theirs"],
            desc = L["Activates this aura group only for auras not cast by you."],
            get = function()
                return sets.onlyTheirs
                end,
            set = function(v)
                sets.onlyTheirs = v
                if v and sets.onlyMine then
                    sets.onlyMine = false
                end
                self:UpdateAllUnitAuras()
            end,
            order = 143,
        },
        ["missing"] = {
            type = "toggle",
            name = L["All missing"],
            desc = L["Activates this aura group only if all of its auras are missing."],
            get = function()
                return sets.allMissing
                end,
            set = function(v)
                sets.allMissing = v
                self:UpdateAllUnitAuras()
            end,
            order = 144,
        },
    }
    
    self:RegisterStatus(status, name, opts, true)
    
    for id, _ in pairs(sets.ids) do
        self:RegisterId(status, id)
    end
    
    for aura, _ in pairs(sets.names) do
        self:RegisterName(status, aura)
    end
end

function GridStatusAurasExt:UnregisterName(status, aura)
    auraList[aura][status] = nil
end

function GridStatusAurasExt:RegisterName(status, aura)
    local auraData = auraList[aura]
    if not auraData then
        auraData = {}
        auraList[aura] = auraData
    end
    
    auraData[status] = {settings = settings[status]}    
    
    local opts = GridStatus.options.args[status].args
    opts[aura] = {
        type = "toggle",
        name = aura,
        desc = aura,
        get = function()
            return settings[status].names[aura]
            end,
        set = function(v)
            settings[status].names[aura] = v
        end,
        order = 160,
    }
    opts.remove.args[aura] = {
        type = "execute",
        name = aura,
        desc = L["Removes this aura."],        
        func = function() self:RemoveNameFromGroup(status, aura) end,
        order = 1,
    }
end

function GridStatusAurasExt:UnregisterId(status, id)
    local aura, rank, icon = GetSpellInfo(id)
    auraList[aura][status] = nil
end

function GridStatusAurasExt:RegisterId(status, id)
    local aura, rank, icon = GetSpellInfo(id)
    local auraData = auraList[aura]
    if not auraData then
        auraData = {}
        auraList[aura] = auraData
    end
    
    auraData[status] = {
        settings = settings[status], 
        icon = icon,
        rank = rank,
    }

    local opts = GridStatus.options.args[status].args
    
    opts[id] = {
        type = "toggle",
        name = GetSpellInfo(id).."("..id..")",
        desc = id,
        get = function()
            return settings[status].ids[id]
            end,
        set = function(v)
            settings[status].ids[id] = v
        end,
        order = 160,
    }
    opts.remove.args[id] = {
        type = "execute",
        name = GetSpellInfo(id).."("..id..")",
        desc = L["Removes this aura."],        
        func = function() self:RemoveIdFromGroup(status, id) end,
        order = 1,
    }
end

function GridStatusAurasExt:RemoveIdFromGroup(status, id)
    local opts = GridStatus.options.args[status].args
    opts[id] = nil
    opts.remove.args[id] = nil
    settings[status].ids[id] = nil
    self:UpdateAllUnitAuras()
end

function GridStatusAurasExt:RemoveNameFromGroup(status, aura)
    local opts = GridStatus.options.args[status].args
    opts[aura] = nil
    opts.remove.args[aura] = nil
    settings[status].names[aura] = nil
    self:UpdateAllUnitAuras()
end

function GridStatusAurasExt:NewAuraIDToGroup(status, id)
    self:RegisterId(status, id)
    settings[status].ids[id] = true
    self:UpdateAllUnitAuras()
end

function GridStatusAurasExt:NewAuraNameToGroup(status, aura)
    self:RegisterName(status, aura)
    settings[status].names[aura] = true
    self:UpdateAllUnitAuras()
end

function GridStatusAurasExt:NewAuraGroup(name)   
    settings.auraGroups[name] = true
    self:RegisterAuraGroup(name, settings.auraGroups[name])
end

function GridStatusAurasExt:OnEnable()
	--self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
    self:RegisterEvent("UNIT_AURA")
    self:RegisterEvent("Grid_UnitJoined")
    
    self:UpdateAllUnitAuras()
end

function GridStatusAurasExt:UpdateAllUnitAuras()
	for guid, unitid in GridRoster:IterateRoster() do
		self:UNIT_AURA(unitid)
	end
end


function GridStatusAurasExt:Grid_UnitJoined(guid, unitid)
	self:UNIT_AURA(unitid)
end

function GridStatusAurasExt:UNIT_AURA(unit)
    local guid = UnitGUID(unit)
	if not GridRoster:IsGUIDInRaid(guid) then
		return
	end
    
    for status, data in pairs(groupList) do
        data.active = false
    end

    local i = 1
    local name, rank, icon, count, debuffType, duration, expirationTime, source, isStealable = UnitDebuff(unit, i)
    while name do
        local aura = auraList[name]
        if aura then
            local isMine = source and UnitIsUnit(source, "player")
            for status, data in pairs(aura) do
                if  not groupList[status].active
                    and (not data.icon or data.icon == icon) 
                    and (not data.rank or data.rank == rank)
                    and (not data.settings.onlyMine or isMine)
                    and (not data.settings.onlyTheirs or not isMine) then
                    
                    groupList[status].active = true
                    groupList[status].settings.name = name
                    groupList[status].settings.icon = icon
                    groupList[status].start = expirationTime - duration
                    groupList[status].duration = duration
                    groupList[status].count = count
                end
            end
        end
        i = i + 1
        name, rank, icon, count, debuffType, duration, expirationTime, source, isStealable = UnitDebuff(unit, i)
    end
    
    local i = 1
    local name, rank, icon, count, debuffType, duration, expirationTime, source, isStealable = UnitBuff(unit, i)
    while name do
        local aura = auraList[name]
        if aura then
            local isMine = source and UnitIsUnit(source, "player")
            for status, data in pairs(aura) do
                if  not groupList[status].active
                    and (not data.icon or data.icon == icon) 
                    and (not data.rank or data.rank == rank)
                    and (not data.settings.onlyMine or isMine)
                    and (not data.settings.onlyTheirs or not isMine) then
                    
                    groupList[status].active = true
                    groupList[status].settings.name = name
                    groupList[status].settings.icon = icon
                    groupList[status].start = expirationTime - duration
                    groupList[status].duration = duration
                    groupList[status].count = count
                end
            end
        end
        i = i + 1
        name, rank, icon, count, debuffType, duration, expirationTime, source, isStealable = UnitBuff(unit, i)
    end
    
    for status, data in pairs(groupList) do
        if data.active == (not data.settings.allMissing) then
            --print(status.." "..(data.settings.name or "nil"))
            self.core:SendStatusGained(guid,
                           status,
                           data.settings.priority,
                           (data.settings.range and 40),
                           data.settings.color,
                           data.settings.name,
                           data.count,
                           nil,
                           data.settings.icon,
                           (data.active and data.start) or nil,
                           (data.active and data.duration) or nil,
                           (data.active and data.count) or nil)
        else
            --print(status.." lost")
            self.core:SendStatusLost(guid, status)
        end
    end
end



function GridStatusAurasExt:BruteForceAuraData(unit, id, aura, isMine, isTheirs, buffsOnly, debuffOnly)
    assert(not(isMine and isTheirs))
    assert(not(buffsOnly and debuffOnly))
    
    local idIcon
    if id then
        _, _, idIcon = GetSpellInfo(id)
    end
    
    if not buffsOnly then
        for i=1,40 do
            local name, rank, icon, count, debuffType, duration, expirationTime, source, isStealable = UnitDebuff(unit, i)
            if name then
                print(name)
                local mine = UnitIsUnit(source, "player")
                if name == aura and (not id or icon == idIcon) and (not isMine or mine) and (not isTheirs or not mine) then
                    return icon, count, duration, expirationTime
                end
            end
        end
    end
    
    if not debuffsOnly then
        for i=1,40 do
            local name, rank, icon, count, debuffType, duration, expirationTime, source, isStealable = UnitBuff(unit, i)
            if name then
                print(name)
                local mine = source and UnitIsUnit(source, "player")
                if name == aura and (not id or icon == idIcon) and (not isMine or mine) and (not isTheirs or not mine) then
                    return icon, count, duration, expirationTime
                end
            end
        end
    end
end 


function GridStatusAurasExt:SendAura(group, spell, guid, icon, count, duration, expirationTime)
    local status = "status_"..group
    local grpSets = settings[status]
    print(guid)
    
    local start = nil
    if duration and expirationTime then
        start = expirationTime - duration
    end
    
    self.core:SendStatusGained(guid,
                           status,
                           grpSets.priority,
                           (grpSets.range and 40),
                           grpSets.color,
                           spell,
                           count,
                           nil,
                           icon,
                           start,
                           duration,
                           count)
end

function GridStatusAurasExt:LostAura(group, guid)
    self.core:SendStatusLost(guid, "status_"..group)
end

function GridStatusAurasExt:AuraApplied(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellId, spellName, spellSchool, auraType)  
    spellId = tostring(spellId)
    for grp, grpData in pairs(settings.auraGroups) do
        if grpData.ids[spellId] or grpData.names[spellName] then
            print("isaura")
            local icon, count, duration, expTime = self:BruteForceAuraData(dstName, spellId, spellName, grpData.onlyMine, grpData.onlyTheirs)
            if icon then 
                grpData.icon = icon
                local auras = grpAuras[grp]
                
                if not auras[spellName] then
                    auras[spellName] = {count = 0}
                end
                if not auras[spellName][spellId] then
                    auras[spellName].count = auras[spellName].count + 1
                    auras.count = auras.count + 1
                end
                auras[spellName][spellId] = true
                
                if grpData.allMissing then
                    self:LostAura(grp, dstGUID)
                else
                    self:SendAura(grp, spellName, dstGUID, icon, count, duration, expTime)
                end
            end
        end
    end
end

function GridStatusAurasExt:AuraRefresh(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellId, spellName, spellSchool, auraType)    
    assert(false)
end

function GridStatusAurasExt:AuraRemoved(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellId, spellName, spellSchool, auraType)  
    spellId = tostring(spellId)
    for grp, grpData in pairs(settings.auraGroups) do     
        local auras = grpAuras[grp][spellName]
        if auras and auras[spellId] and (grpData.ids[spellId] or grpData.names[spellName]) then
            local icon, count, duration, expTime = self:BruteForceAuraData(dstName, spellId, spellName, grpData.onlyMine, grpData.onlyTheirs)
            
            if icon then     
                if grpData.allMissing then
                    self:LostAura(grp, dstGUID)
                else
                    self:SendAura(grp, spellName, dstGUID, icon, count, duration, expTime)
                end
            else                
                auras[spellId] = nil
                auras.count = auras.count - 1
                grpAuras[grp].count = grpAuras[grp].count - 1
                

                if grpAuras[grp].count <= 0 then
                    assert(grpAuras[grp].count == 0)
                    local status = "status_"..grp
                    if grpData.allMissing then
                        self:SendAura(grp, spellName, dstGUID, grpData.icon)
                    else
                        self:LostAura(grp, dstGUID)
                    end
                else
                    for name, auras in pairs(grpAuras[grp]) do
                        for id, _ in pairs(auras) do
                            if name ~= "count" then
                                icon, count, duration, expTime = self:BruteForceAuraData(dstName, id, name, grpData.onlyMine, grpData.onlyTheirs)
                                assert(icon)
                                if grpData.allMissing then
                                    self:LostAura(grp, dstGUID)
                                else
                                    self:SendAura(grp, name, dstGUID, icon, count, duration, expTime)
                                end
                                break
                            end
                        end
                    end
                end
            end
        end
    end
end


local EventParse = {}
EventParse["SPELL_AURA_APPLIED"] = GridStatusAurasExt.AuraApplied
EventParse["SPELL_AURA_REFRESH"] = GridStatusAurasExt.AuraApplied
EventParse["SPELL_AURA_REMOVED"] = GridStatusAurasExt.AuraRemoved 
EventParse["SPELL_AURA_BROKEN"] = GridStatusAurasExt.AuraRemoved
EventParse["SPELL_AURA_BROKEN_SPELL"] = GridStatusAurasExt.AuraRemoved

local bit_band = _G.bit.band

local function isInPartyOrRaidOrMe(unitFlags)
    return  bit_band(unitFlags, COMBATLOG_OBJECT_AFFILIATION_RAID) == COMBATLOG_OBJECT_AFFILIATION_RAID or 
            bit_band(unitFlags, COMBATLOG_OBJECT_AFFILIATION_PARTY) == COMBATLOG_OBJECT_AFFILIATION_PARTY or 
            bit_band(unitFlags, COMBATLOG_OBJECT_AFFILIATION_MINE) == COMBATLOG_OBJECT_AFFILIATION_MINE 
end



function GridStatusAurasExt:COMBAT_LOG_EVENT_UNFILTERED(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, ...)   
    if not (isInPartyOrRaidOrMe(dstFlags)) then
        return
    end
	local parsefunc = EventParse[eventtype]
	
    --print("Event: "..eventtype)
	if parsefunc then  
		parsefunc(self, timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, ...)
	end
end
